from dataclasses import dataclass
from os import sep
from pathlib import Path
from subprocess import run
from typing import Dict, List, Optional

import yaml
from cleo.io.io import IO
from poetry.console.application import Application
from poetry.poetry import Poetry


def find_root_dir(poetry: Poetry) -> Path:
    return poetry.pyproject.path.parent


@dataclass
class AwsLambda:
    name: str
    path: Path


class Sam:
    def __init__(self, sam_exec: str, template: Path):
        self.exec = sam_exec
        self.template_path = template
        self.template = self._parse_sam_template()
        self.lambdas = self._get_aws_lambdas()

    def _get_aws_lambdas(self) -> List[AwsLambda]:
        resources = self.template["Resources"]
        lambdas = {
            resource: {
                **self.template.get("Globals", {}).get("Function", {}),
                **resources[resource]["Properties"],
            }
            for resource, param in resources.items()
            if param["Type"] == "AWS::Serverless::Function"
        }
        return [
            AwsLambda(
                name=resource,
                path=Path(param["Handler"].replace(".", sep)).parent.parent,
            )
            for resource, param in lambdas.items()
            if param.get("Runtime", "").lower().startswith("python")
        ]

    def _parse_sam_template(self) -> Dict:
        yaml.SafeLoader.add_multi_constructor("!", lambda *args: None)
        return yaml.safe_load(self.template_path.read_text(encoding="utf-8"))

    def invoke_sam_build(self, build_dir: str, params: Optional[List[str]] = None):
        def_params = ["--template", str(self.template_path), "--build-dir", build_dir]
        if not params:
            params = []
        params.extend(def_params)

        return run(
            [self.exec, "build"] + params,
            text=True,
            encoding="utf-8",
            capture_output=True,
            check=False,
        )



import sys
from shlex import quote
from subprocess import PIPE, check_call
from typing import Dict

from poetry_aws_sam.aws import AwsLambda, Sam, find_root_dir
from poetry_aws_sam.export import ExportLock

SAM_BUILD_DIR_NAME = ".aws-sam/build"


class AwsBuilder:
    def __init__(self, options, poetry, io):
        self.config = options
        self._io = io
        self.poetry = poetry

    @property
    def root_dir(self):
        return find_root_dir(self.poetry)

    def get_version_api(self) -> Dict:
        return {"standard": self.build_standard}

    def abort(self, message: str = "", code: int = 1) -> None:
        """
        Terminate the program with the given return code.
        """
        self._io.write_error_line(message)

        sys.exit(code)

    @property
    def sam_build_location(self):
        return self.root_dir / SAM_BUILD_DIR_NAME

    def build_lambda(self, aws_lambda: AwsLambda) -> None:
        build_dir = self.sam_build_location / aws_lambda.name
        target = build_dir / aws_lambda.path
        requirements_file = target / "requirements.txt"
        requirements_file.parent.mkdir(exist_ok=True, parents=True)
        _ = ExportLock(self.config, self.poetry, self._io).handle(requirements_file)

        check_call(
            [
                sys.executable,
                "-m",
                "pip",
                "install",
                "--upgrade",
                "--disable-pip-version-check",
                "--no-python-version-warning",
                "-r",
                quote(str(requirements_file)),
                "-t",
                quote(str(build_dir)),
            ],
            stdout=PIPE,
            stderr=PIPE,
            shell=False,
        )

    def build_standard(self) -> int:
        try:
            sam = Sam(sam_exec="sam", template=self.root_dir / self.config("sam-template"))
        except AttributeError:
            self.abort(
                "Unsupported type for a 'CodeUri' or 'Handler'. Only string is supported."
                "Functions !Sub, !Ref and others are not supported yet. "
            )
            raise
        self._io.write_line("Building lambda functions ...")
        build_dir = str(self.sam_build_location)
        result = sam.invoke_sam_build(build_dir=build_dir, params=None)
        if result.returncode != 0:
            self._io.write_error_line(result.stderr)
            self.abort("SAM build failed!")

        for aws_lambda in sam.lambdas:
            self._io.write_line(f"{aws_lambda.name} ...")
            self.build_lambda(aws_lambda=aws_lambda)
            self._io.write_line("success")

        self._io.write_line("Build successfull 🚀")
        return 0



from collections import defaultdict
from collections.abc import Iterable
from pathlib import Path

from cleo.io.outputs.output import Verbosity
from packaging.utils import NormalizedName, canonicalize_name
from poetry.console.exceptions import GroupNotFound
from poetry.core.packages.dependency_group import MAIN_GROUP
from poetry_plugin_export.exporter import Exporter

FORMAT_REQUIREMENTS_TXT = "requirements.txt"


class ExportLock:
    """
    Taken from ExportCommand of poetry-plugin-export.
    https://github.com/python-poetry/poetry-plugin-export/blob/2d370be73a9830a30218ebaa6e0366d974208706/src/poetry_plugin_export/command.py#L19
    """

    def __init__(self, options, poetry, io):
        self.config = options
        self._io = io
        self.poetry = poetry

    def _validate_group_options(self, group_options: dict[str, set[str]]) -> None:
        """
        Raises an error if it detects that a group is not part of pyproject.toml
        """
        invalid_options = defaultdict(set)
        for opt, groups in group_options.items():
            for group in groups:
                if not self.poetry.package.has_dependency_group(group):
                    invalid_options[group].add(opt)
        if invalid_options:
            message_parts = []
            for group in sorted(invalid_options):
                opts = ", ".join(f"<fg=yellow;options=bold>--{opt}</>" for opt in sorted(invalid_options[group]))
                message_parts.append(f"{group} (via {opts})")
            raise GroupNotFound(f"Group(s) not found: {', '.join(message_parts)}")

    def line_error(
        self,
        text: str,
        style: str | None = None,
        verbosity: Verbosity = Verbosity.NORMAL,
    ) -> None:
        """
        Write a string as information output to stderr.
        """
        styled = f"<{style}>{text}</>" if style else text

        self._io.write_error_line(styled, verbosity)

    @property
    def activated_groups(self) -> set[str]:
        groups = {}

        for key in {"with", "without", "only"}:
            groups[key] = {group.strip() for groups in self.config(key, "") for group in groups.split(",")}
        self._validate_group_options(groups)

        for opt, new, group in [
            ("no-dev", "only", MAIN_GROUP),
            ("dev", "with", "dev"),
        ]:
            if self._io.input.has_option(opt) and self.config(opt):
                self.line_error(
                    f"<warning>The `<fg=yellow;options=bold>--{opt}</>` option is"
                    f" deprecated, use the `<fg=yellow;options=bold>--{new} {group}</>`"
                    " notation instead.</warning>"
                )
                groups[new].add(group)

        if groups["only"] and (groups["with"] or groups["without"]):
            self.line_error(
                "<warning>The `<fg=yellow;options=bold>--with</>` and "
                "`<fg=yellow;options=bold>--without</>` options are ignored when used"
                " along with the `<fg=yellow;options=bold>--only</>` option."
                "</warning>"
            )

        return groups["only"] or self.default_groups.union(groups["with"]).difference(groups["without"])

    @property
    def non_optional_groups(self) -> set[str]:
        # method only required for poetry <= 1.2.0-beta.2.dev0
        return {MAIN_GROUP}

    @property
    def default_groups(self) -> set[str]:
        return {MAIN_GROUP}

    def handle(self, requirements_file: Path) -> int:
        fmt = FORMAT_REQUIREMENTS_TXT

        if not Exporter.is_format_supported(fmt):
            raise ValueError(f"Invalid export format: {fmt}")

        locker = self.poetry.locker
        if not locker.is_locked():
            self.line_error("<comment>The lock file does not exist. Locking.</comment>")
            options = []
            if self._io.is_debug():
                options.append(("-vvv", None))
            elif self._io.is_very_verbose():
                options.append(("-vv", None))
            elif self._io.is_verbose():
                options.append(("-v", None))

            self.call("lock", " ".join(options))  # type: ignore[arg-type]

        if not locker.is_fresh():
            self.line_error(
                "<warning>"
                "Warning: poetry.lock is not consistent with pyproject.toml. "
                "You may be getting improper dependencies. "
                "Run `poetry lock [--no-update]` to fix it."
                "</warning>"
            )

        # Checking extras
        if self.config("extras") and self.config("all-extras"):
            self.line_error(
                "<error>You cannot specify explicit"
                " `<fg=yellow;options=bold>--extras</>` while exporting"
                " using `<fg=yellow;options=bold>--all-extras</>`.</error>"
            )
            return 1

        extras: Iterable[NormalizedName]
        if self.config("all-extras"):
            extras = self.poetry.package.extras.keys()
        else:
            extras = {canonicalize_name(extra) for extra_opt in self.config("extras") for extra in extra_opt.split()}
            invalid_extras = extras - self.poetry.package.extras.keys()
            if invalid_extras:
                raise ValueError(f"Extra [{', '.join(sorted(invalid_extras))}] is not specified.")

        exporter = Exporter(self.poetry, self._io)
        exporter.only_groups(list(self.activated_groups))
        exporter.with_extras(list(extras))
        exporter.with_hashes(not self.config("without-hashes"))
        exporter.with_credentials(self.config("with-credentials"))
        exporter.with_urls(not self.config("without-urls"))
        exporter.export(fmt, Path.cwd(), str(requirements_file))

        return 0
